# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


import logging
from dataclasses import dataclass
from typing import Optional

import torch
from torch import nn

from xformers.components.attention import Attention, AttentionConfig, register_attention

_mask_type_warning = True

_use_triton = torch.cuda.is_available()
if _use_triton:
    try:
        from triton.ops.blocksparse import matmul as blocksparse_matmul
        from triton.ops.blocksparse import softmax as blocksparse_softmax

        from xformers.triton.softmax import MaskType

    except ImportError as e:
        logging.warning(
            f"Triton is not available: {e}.\nBlockSparse attention will not be available"
        )
        _use_triton = False


if _use_triton:

    @dataclass
    class BlockSparseAttentionConfig(AttentionConfig):
        layout: torch.Tensor  # The dimensions of the random features
        block_size: int
        dropout: float
        num_heads: int

    @register_attention("blocksparse", BlockSparseAttentionConfig)
    class BlockSparseAttention(Attention):
        r"""
        Thin wrap over the Triton blocksparse computations. The sparsity pattern is determined through the layout.

        .. warning: the layout is assumed to have the dimensions [heads, seq, seq].
            If some dimensions are missing, we assume that the same layout is to be used across heads.

        .. note: it is possible to pass a specific per batch mask in the forward call,
            but this will not lead to any speed up.
            Any constant sparsity pattern is better passed through the layout parameter.
        """

        def __init__(
            self,
            layout: torch.Tensor,
            block_size: int = 16,
            dropout: float = 0.0,
            num_heads: int = 1,  # optional, used to adapt the layout if in need
            *args,
            **kwargs,
        ):
            if layout.dim() == 2:
                logging.warning(
                    "The layout passed is lacking a head dimension and a batch dimension"
                )
                logging.warning(
                    "Now assuming that the same layout is to be used across all heads"
                )
                layout = layout.unsqueeze(0).expand(num_heads, -1, -1)
                logging.warning(f"New layout dimensions: {layout.shape}")

            assert block_size >= 16, "Minimum block size is 16, for now at least"

            super().__init__()
            self.attn_drop = nn.Dropout(dropout, inplace=False)

            # Pure blocksparse data
            self.layout = layout
            self.block_size = block_size

            # blocksparse operators
            self.sparse_dot_sdd = blocksparse_matmul(
                self.layout,
                self.block_size,
                "sdd",
                trans_a=False,
                trans_b=True,
            )
            self.sparse_dot_dsd = blocksparse_matmul(
                self.layout,
                self.block_size,
                "dsd",
                trans_a=False,
                trans_b=False,
            )
            self.sparse_softmax = blocksparse_softmax(self.layout, self.block_size)

            # make sure that the head dimension is not folded down with the batch
            self.requires_head_dimension = True

            # key padding mask and attention mask must be passed in separately
            self.requires_separate_masks = True

        def forward(
            self,
            q: torch.Tensor,
            k: torch.Tensor,
            v: torch.Tensor,
            att_mask: Optional[torch.Tensor] = None,
            key_padding_mask: Optional[torch.Tensor] = None,
            scale: float = 1.0,
            *args,
            **kwargs,
        ) -> torch.Tensor:

            # NOTE:
            # The attention mask will be taken into account when computing the softmax
            # meaning that non-masked values which are present in the initial blocksparse layout will be computed.
            # If blocks are to be constantly masked, better perf would thus be reached by signalling them out in the
            # initial attention setup

            if att_mask is not None and att_mask.dtype != q.dtype:
                global _mask_type_warning
                if _mask_type_warning:
                    logging.warning(
                        "Attention mask has to be multiplicative. Fixing that but this slows things down"
                    )
                    _mask_type_warning = False  # Only warn once
                att_mask = att_mask.to(q.dtype)

            assert (
                att_mask is None or att_mask.dim() == 2
            ), "The attention mask is constant across heads, expected dimensions are [seq x seq]"

            # Self-attend: (B, nh, S, hs) x (B, nh, hs, S) -> (B, nh, S, S)
            # When the computations are block sparse, the matrix types change along the way:
            # - (sparse) attention matrix = (dense) Kt * (dense) Q
            assert (
                q.shape[-2] == k.shape[-2]
            ), "Blocksparse requires the same dimensions for K and Q for now"
            sparse_att_mat = self.sparse_dot_sdd(q, k)

            # - softmax on the sparse attention matrix
            sparse_att_mat = self.sparse_softmax(
                sparse_att_mat,
                scale=scale,
                key_padding_mask=key_padding_mask,
                attn_mask=att_mask,
                key_padding_mask_mode=MaskType.MUL,
                attn_mask_mode=MaskType.MUL,
            )

            # - then (dense) attention is (sparse) attention matrix * dense (value)
            a = self.sparse_dot_dsd(sparse_att_mat, v)

            return a



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


from collections import namedtuple
from dataclasses import asdict, dataclass
from typing import Optional, Tuple

import torch
import torch.nn as nn
from torch.nn.init import constant_, xavier_uniform_

from xformers.components.attention import Attention

InProjParams = namedtuple("InProjParams", ["in_features", "out_features", "bias"])


class InProjContainer(nn.Module):
    """
    Handle all the input projections in one go, opportunistically fuse some operations.

    CREDITS: Inspired by https://github.com/pytorch/text/blob/master/torchtext/nn/modules/multiheadattention.py
    and the MultiHeadAttention implementation from PyTorch
    """

    def __init__(
        self,
        query_proj_params: InProjParams,
        key_proj_params: Optional[InProjParams],
        value_proj_params: Optional[InProjParams],
    ):

        super().__init__()

        assert (
            query_proj_params.in_features == query_proj_params.out_features
        ), "We assume in_features == out_features for queries, please provide your projection if this is not the case"

        # If nothing is specified for key and value, use the same as query
        if key_proj_params is None:
            key_proj_params = query_proj_params

        if value_proj_params is None:
            value_proj_params = query_proj_params

        # Catch a beneficial case, if Q,K,V dimensions are the same
        self.same_dimensions = (
            query_proj_params.in_features == key_proj_params.in_features
            and value_proj_params.in_features == key_proj_params.in_features
        )

        self.out_features = query_proj_params.out_features

        # - handle all the weights
        if self.same_dimensions:
            # We can use a single weight and bias buffer, which will speed up self attention
            self.in_proj_weight = nn.Parameter(
                torch.empty((3 * self.out_features, self.out_features))
            )
            self.register_parameter("q_proj_weight", None)
            self.register_parameter("k_proj_weight", None)
            self.register_parameter("v_proj_weight", None)
        else:
            # The dimensions are different, use seperate buffers
            self.q_proj_weight = nn.Parameter(
                torch.empty((self.out_features, query_proj_params.in_features))
            )
            self.k_proj_weight = nn.Parameter(
                torch.empty((self.out_features, key_proj_params.in_features))
            )
            self.v_proj_weight = nn.Parameter(
                torch.empty((self.out_features, value_proj_params.in_features))
            )
            self.register_parameter("in_proj_weight", None)

        # - handle all the inputs
        if query_proj_params.bias:
            self.in_proj_bias = nn.Parameter(torch.empty(3 * self.out_features))
        else:
            self.register_parameter("in_proj_bias", None)

        # - multi-head attention specific init for the weights and biases
        self._reset_parameters()

    def _reset_parameters(self):
        if self.in_proj_weight is not None:
            xavier_uniform_(self.in_proj_weight)
        else:
            xavier_uniform_(self.q_proj_weight)
            xavier_uniform_(self.k_proj_weight)
            xavier_uniform_(self.v_proj_weight)

        if self.in_proj_bias is not None:
            constant_(self.in_proj_bias, 0.0)

    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        if self.in_proj_weight is not None:
            if id(query) == id(key):
                # Self attention, get all the projected values at once
                # we compute everything transposed, so that q,k,v stay contiguous after splitting
                qkv = query @ self.in_proj_weight.transpose(-2, -1)

                if self.in_proj_bias is not None:
                    qkv += self.in_proj_bias

                q, k, v = map(
                    lambda x: x.contiguous(),
                    qkv.split(self.out_features, dim=-1),
                )
                return q, k, v

            else:
                # Not self attention
                # - bias free projection
                projections = self.in_proj_weight.split(self.out_features, dim=0)
                q, k, v = map(
                    lambda x, y: x @ y.transpose(1, 0), [query, key, value], projections
                )

                # - optionally add bias
                if self.in_proj_bias is not None:
                    biases = self.in_proj_bias.split(self.out_features, dim=0)
                    q, k, v = map(lambda x, y: x + y, [q, k, v], biases)

                return q, k, v

        # We have a weight per input, but share a bigger bias buffer
        assert (
            self.q_proj_weight is not None
            and self.k_proj_weight is not None
            and self.v_proj_weight is not None
        )

        # - bias free projection
        q, k, v = map(
            lambda x, y: x @ y.transpose(1, 0),
            [query, key, value],
            [self.q_proj_weight, self.k_proj_weight, self.v_proj_weight],
        )

        # - optionally add bias
        if self.in_proj_bias is not None:
            biases = self.in_proj_bias.split(self.out_features, dim=0)
            q, k, v = map(lambda x, y: x + y, [q, k, v], biases)

        return q, k, v


@dataclass
class MultiHeadDispatchConfig:
    dim_model: int
    residual_dropout: float
    num_heads: int
    attention: Attention
    bias: bool
    dim_key: Optional[int]
    dim_value: Optional[int]
    in_proj_container: Optional[InProjContainer]
    use_separate_proj_weight: Optional[bool]
    out_proj: Optional[nn.Module]


# Move head forward and fold into batch dim. dimensions become (B * nh, S, hs)
def _fold_heads(t: torch.Tensor, B: int, S: int, H: int, Hs: int):
    return t.view(B, S, H, Hs).transpose(1, 2).flatten(start_dim=0, end_dim=1)


def _split_heads(t: torch.Tensor, B: int, S: int, H: int, Hs: int):
    return t.view(B, S, H, Hs).transpose(1, 2)


class MultiHeadDispatch(nn.Module):
    """
    A multi-head masked self-attention dispatch mechanism, with a projection at the end,
    following the architecture proposed in `Attention is all you need`_, Vaswani et al.

    The actual attention mechanism can vary, as well as the projections.
    This can be used to wrap the proposed attention mechanisms and make them multi-head aware,
    but it is optional.

    .. _`Attention is all you need`: https://arxiv.org/abs/1706.03762v5
    """

    def __init__(
        self,
        dim_model: int,
        residual_dropout: float,
        num_heads: int,
        attention: Attention,
        bias: bool = True,
        dim_key: Optional[int] = None,
        dim_value: Optional[int] = None,
        in_proj_container: Optional[InProjContainer] = None,
        use_separate_proj_weight: Optional[bool] = False,
        out_proj: Optional[nn.Module] = None,
        *args,
        **kwargs,
    ):
        super().__init__()

        assert (
            dim_model % num_heads == 0
        )  # static preset for now, each head works on 1/d the embeddings, could be relaxed
        assert num_heads > 0

        # Popular default is that all latent dimensions are the same
        dim_key, dim_value = map(lambda x: x if x else dim_model, (dim_key, dim_value))

        self.num_heads = num_heads
        self.dim_k = dim_key // num_heads
        self.dim_value = dim_value
        self.dim_model = dim_model
        self.attention = attention

        # key, query, value projections for all heads
        # critical options are
        # - are we sharing weights ?
        # - are we adding biases, and if yes are they shared ?
        if attention.requires_input_projection:
            self.in_proj_container = (
                in_proj_container
                if in_proj_container is not None
                else InProjContainer(
                    query_proj_params=InProjParams(dim_model, dim_key, bias=bias),
                    key_proj_params=InProjParams(dim_model, dim_key, bias=bias)
                    if use_separate_proj_weight
                    else None,
                    value_proj_params=InProjParams(dim_model, dim_value, bias=bias)
                    if use_separate_proj_weight
                    else None,
                )
            )

        # Regularization
        self.resid_drop = nn.Dropout(residual_dropout, inplace=False)

        # Output projection
        self.proj = out_proj if out_proj else nn.Linear(dim_model, dim_model, bias=bias)
        if isinstance(self.proj, nn.Linear) and self.proj.bias is not None:
            constant_(self.proj.bias, 0.0)

    def _check(self, t, name):
        assert (
            t.shape[2] % self.dim_k == 0
        ), f"the {name} embeddings need to be divisible by the number of heads"

    def forward(
        self,
        query: torch.Tensor,
        key: Optional[torch.Tensor] = None,
        value: Optional[torch.Tensor] = None,
        att_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """
        Expected input dimensions are [batch size, sequence length, embed dim]
        Output dimensions are [batch size, sequence length, embed dim]
        """

        if key is None:
            key = query
        if value is None:
            value = query

        # Check the dimensions properly
        self._check(query, "query")
        self._check(value, "value")
        self._check(key, "key")

        B, S_Q, _ = query.size()  # Batch x Sequence x Embedding (latent)
        _, S_K, _ = key.size()  # K, Q's sequence length could differ

        # Calculate query, key, values for all heads in batch
        if self.attention.requires_input_projection:
            q, k, v = self.in_proj_container(query=query, key=key, value=value)
        else:
            k, q, v = key, query, value

        # Reshape k/q/v to either expose the heads, or fold the head dimension into the batch
        reshape_fn = (
            _split_heads if self.attention.requires_head_dimension else _fold_heads
        )

        k = reshape_fn(k, B, S_K, self.num_heads, self.dim_k)
        q = reshape_fn(q, B, S_Q, self.num_heads, self.dim_k)
        v = reshape_fn(v, B, S_K, self.num_heads, self.dim_k)

        # Self-attend
        y = self.attention(q=q, k=k, v=v, att_mask=att_mask)

        # Re-assemble all head outputs side by side
        y = (
            y.view(B, self.num_heads, S_Q, self.dim_k)
            .transpose(1, 2)
            .flatten(start_dim=2, end_dim=3)
        )

        # Output projection, dropout and good to go
        y = self.resid_drop(self.proj(y))

        # Return the same sequence size as the input
        return y

    @classmethod
    def from_config(cls, config: MultiHeadDispatchConfig):
        # Generate the class inputs from the config
        fields = asdict(config)

        # Skip all Nones so that default values are used
        fields = {k: v for k, v in fields.items() if v is not None}

        return cls(**fields)



# Copyright 2021 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generators for custom listops tasks."""

import csv
import random

from absl import app
from absl import flags
import numpy as np
import tensorflow.compat.v1 as tf

flags.DEFINE_string(
    'task', default='basic',
    help='Name of task to create.')
flags.DEFINE_integer(
    'num_train_samples', default=96000,
    help=('Number of train samples.'))
flags.DEFINE_integer(
    'num_valid_samples', default=2000,
    help=('Number of test samples.'))
flags.DEFINE_integer(
    'num_test_samples', default=2000,
    help=('Number of test samples.'))
flags.DEFINE_integer(
    'max_depth', default=10,
    help=('maximum tree depth of training sequences.'))
flags.DEFINE_integer(
    'max_args', default=10,
    help=('maximum number of arguments per operator in training sequences.'))
flags.DEFINE_integer(
    'max_length', default=2000,
    help=('maximum length per sequence in training sequences.'))
flags.DEFINE_integer(
    'min_length', default=500,
    help=('minimum length per sequence in training sequences.'))
flags.DEFINE_string(
    'output_dir', default='output_dir',
    help='Directory to output files.')

FLAGS = flags.FLAGS

MIN = '[MIN'
MAX = '[MAX'
MED = '[MED'
FIRST = '[FIRST'
LAST = '[LAST'
SUM_MOD = '[SM'
END = ']'

OPERATORS = [MIN, MAX, MED, SUM_MOD]  # , FIRST, LAST]
VALUES = range(10)

VALUE_P = 0.25


def generate_tree(depth, max_depth, max_args):
  """Generate tree-like equations.

  Args:
    depth: current depth of the node, int.
    max_depth: maximum depth of the tree, int.
    max_args: maximum number of arguments per operator, int.

  Returns:
    The root node of a tree structure.
  """
  if depth < max_depth:
    r = random.random()
  else:
    r = 1

  if r > VALUE_P:
    value = random.choice(VALUES)
    return value, 1
  else:
    length = 2
    num_values = random.randint(2, max_args)
    values = []
    for _ in range(num_values):
      sub_t, sub_l = generate_tree(depth + 1, max_depth, max_args)
      values.append(sub_t)
      length += sub_l

    op = random.choice(OPERATORS)
    t = (op, values[0])
    for value in values[1:]:
      t = (t, value)
    t = (t, END)
  return t, length


def to_string(t, parens=True):
  if isinstance(t, str):
    return t
  elif isinstance(t, int):
    return str(t)
  else:
    if parens:
      return '( ' + to_string(t[0]) + ' ' + to_string(t[1]) + ' )'


def to_value(t):
  """Compute the output of equation t.

  Args:
    t: a tree structure that represents equation t, list.

  Returns:
    The result of equation t, int.
  """
  if not isinstance(t, tuple):
    return t
  l = to_value(t[0])
  r = to_value(t[1])
  if l in OPERATORS:  # Create an unsaturated function.
    return (l, [r])
  elif r == END:  # l must be an unsaturated function.
    if l[0] == MIN:
      return min(l[1])
    elif l[0] == MAX:
      return max(l[1])
    elif l[0] == FIRST:
      return l[1][0]
    elif l[0] == LAST:
      return l[1][-1]
    elif l[0] == MED:
      return int(np.median(l[1]))
    elif l[0] == SUM_MOD:
      return np.sum(l[1]) % 10
  elif isinstance(l, tuple):
    # We've hit an unsaturated function and an argument.
    return (l[0], l[1] + [r])


def write_to_file(data, fp):
  """Write to file output."""
  tf.logging.info(type(data))
  tf.logging.info('Writing {} samples to {}'.format(len(data), fp + '.tsv'))
  with tf.io.gfile.GFile(fp + '.tsv', 'w+') as f:
    writer = csv.writer(f, delimiter='\t')
    writer.writerow(['Source', 'Target'])
    writer.writerows(data)


def main(argv):
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')

  tf.logging.info('Start dataset construction')

  data = set()
  num_samples = FLAGS.num_train_samples \
      + FLAGS.num_test_samples + FLAGS.num_valid_samples
  while len(data) < num_samples:
    tree, length = generate_tree(1, FLAGS.max_depth, FLAGS.max_args)
    if length > FLAGS.min_length and length < FLAGS.max_length:
      data.add(tree)
      if len(data) % 1000 == 0:
        tf.logging.info('Processed {}'.format(len(data)))
        print('Processed {}'.format(len(data)))
  train = []
  for example in data:
    train.append([to_string(example), to_value(example)])

  tf.logging.info('Finished running dataset construction')

  val = train[FLAGS.num_train_samples:]
  test = val[FLAGS.num_valid_samples:]
  val = val[:FLAGS.num_valid_samples]
  train = train[:FLAGS.num_train_samples]

  tf.logging.info('Dataset size: %d/%d/%d' % (len(train), len(val), len(test)))

  write_to_file(train, FLAGS.output_dir + '/{}_train'.format(FLAGS.task))
  write_to_file(val, FLAGS.output_dir + '/{}_val'.format(FLAGS.task))
  write_to_file(test, FLAGS.output_dir + '/{}_test'.format(FLAGS.task))
  tf.logging.info('Finished writing all to file')


if __name__ == '__main__':
  app.run(main)



"""Base Configuration."""
# Copyright 2021 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import ml_collections

NUM_EPOCHS = 100
TRAIN_EXAMPLES = 160000
VALID_EXAMPLES = 20000


def get_config():
  """Get the default hyperparameter configuration."""
  config = ml_collections.ConfigDict()
  config.batch_size = 32
  config.eval_frequency = TRAIN_EXAMPLES // config.batch_size
  config.num_train_steps = (TRAIN_EXAMPLES // config.batch_size) * NUM_EPOCHS
  config.num_eval_steps = VALID_EXAMPLES // config.batch_size
  config.weight_decay = 0.
  config.grad_clip_norm = None

  config.save_checkpoints = True
  config.restore_checkpoints = True
  config.checkpoint_freq = (TRAIN_EXAMPLES //
                            config.batch_size) * NUM_EPOCHS // 2
  config.random_seed = 0

  config.learning_rate = .001
  config.factors = 'constant * linear_warmup * cosine_decay'
  config.warmup = (TRAIN_EXAMPLES // config.batch_size) * 1
  config.steps_per_cycle = (TRAIN_EXAMPLES // config.batch_size) * NUM_EPOCHS

  # model params
  config.model = ml_collections.ConfigDict()
  config.model.emb_dim = 32
  config.model.num_heads = 4
  config.model.num_layers = 2
  config.model.qkv_dim = 32
  config.model.mlp_dim = 32
  config.model.dropout_rate = 0.1
  config.model.attention_dropout_rate = 0.1
  config.model.classifier_pool = 'CLS'
  config.model.learn_pos_emb = True

  config.trial = 0  # dummy for repeated runs.
  return config



# Copyright 2021 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""TFDS builder for pathfinder challenge."""

import os

import tensorflow as tf
import tensorflow_datasets as tfds



class Pathfinder32(tfds.core.BeamBasedBuilder):
  """Pathfinder TFDS builder (where the resolution is 32).

  The data for this dataset was generated using the script in
  https://github.com/drewlinsley/pathfinder with the default parameters, while
  followings being customized:
  ```
    args.paddle_margin_list = [1]
    args.window_size = [32, 32]
    args.padding= 1
    args.paddle_length = 2
    args.marker_radius = 1.5
    args.contour_length = 14
    args.paddle_thickness = 0.5
    args.antialias_scale = 2
    args.seed_distance= 7
    args.continuity = 1.0
    args.distractor_length = args.contour_length // 3
    args.num_distractor_snakes = 20 // args.distractor_length
    args.snake_contrast_list = [2]
    args.paddle_contrast_list = [0.75]
  ```
  """

  VERSION = tfds.core.Version('1.0.0')

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=('This is a builder for pathfinder challenge dataset'),
        features=tfds.features.FeaturesDict({
            'image': tfds.features.Image(),
            'label': tfds.features.ClassLabel(num_classes=2)
        }),
        supervised_keys=('image', 'label'),
        homepage='',
        citation="""@inproceedings{
                    Kim*2020Disentangling,
                    title={Disentangling neural mechanisms for perceptual grouping},
                    author={Junkyung Kim* and Drew Linsley* and Kalpit Thakkar and Thomas Serre},
                    booktitle={International Conference on Learning Representations},
                    year={2020},
                    url={https://openreview.net/forum?id=HJxrVA4FDS}
                    }""",
    )

  def _split_generators(self, dl_manager):
    """Downloads the data and defines the splits."""

    return [
        tfds.core.SplitGenerator(
            name='easy', gen_kwargs={'file_pattern': 'curv_baseline'}),
        tfds.core.SplitGenerator(
            name='intermediate',
            gen_kwargs={'file_pattern': 'curv_contour_length_9'}),
        tfds.core.SplitGenerator(
            name='hard', gen_kwargs={'file_pattern': 'curv_contour_length_14'})
    ]

  def _build_pcollection(self, pipeline, file_pattern):
    """Generate examples as dicts."""
    beam = tfds.core.lazy_imports.apache_beam

    def _generate_examples(file_path):
      """Read the input data out of the source files."""
      example_id = 0
      meta_examples = tf.io.read_file(file_path).numpy().decode('utf-8').split(
          '\n')[:-1]
      print(meta_examples)
      for m_example in meta_examples:
        m_example = m_example.split(' ')
        image_path = os.path.join(ORIGINAL_DATA_DIR_32, file_pattern,
                                  m_example[0], m_example[1])
        example_id += 1
        yield '_'.join([m_example[0], m_example[1],
                        str(example_id)]), {
                            'image': image_path,
                            'label': int(m_example[3]),
                        }

    meta_file_pathes = tf.io.gfile.glob(
        os.path.join(ORIGINAL_DATA_DIR_32, file_pattern, 'metadata/*.npy'))
    print(len(meta_file_pathes))
    return (pipeline
            | 'Create' >> beam.Create(meta_file_pathes)
            | 'Generate' >> beam.ParDo(_generate_examples))


class Pathfinder64(tfds.core.BeamBasedBuilder):
  """Pathfinder TFDS builder (where the resolution is 64).

  The data for this dataset was generated using the script in
  https://github.com/drewlinsley/pathfinder with the default parameters, while
  followings being customized:
  ```
    args.padding = 1
    args.antialias_scale = 4
    args.paddle_margin_list = [1]
    args.seed_distance = 12
    args.window_size = [64,64]
    args.marker_radius = 2.5
    args.contour_length = 14
    args.paddle_thickness = 1
    args.antialias_scale = 2
    args.continuity = 1.8  # from 1.8 to 0.8, with steps of 66%
    args.distractor_length = args.contour_length / 3
    args.num_distractor_snakes = 22 / args.distractor_length
    args.snake_contrast_list = [0.8]
  ```
  """

  VERSION = tfds.core.Version('1.0.0')

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=('This is a builder for pathfinder challenge dataset'),
        features=tfds.features.FeaturesDict({
            'image': tfds.features.Image(),
            'label': tfds.features.ClassLabel(num_classes=2)
        }),
        supervised_keys=('image', 'label'),
        homepage='',
        citation="""@inproceedings{
                    Kim*2020Disentangling,
                    title={Disentangling neural mechanisms for perceptual grouping},
                    author={Junkyung Kim* and Drew Linsley* and Kalpit Thakkar and Thomas Serre},
                    booktitle={International Conference on Learning Representations},
                    year={2020},
                    url={https://openreview.net/forum?id=HJxrVA4FDS}
                    }""",
    )

  def _split_generators(self, dl_manager):
    """Downloads the data and defines the splits."""

    return [
        tfds.core.SplitGenerator(
            name='easy', gen_kwargs={'file_pattern': 'curv_baseline'}),
        tfds.core.SplitGenerator(
            name='intermediate',
            gen_kwargs={'file_pattern': 'curv_contour_length_9'}),
        tfds.core.SplitGenerator(
            name='hard', gen_kwargs={'file_pattern': 'curv_contour_length_14'})
    ]

  def _build_pcollection(self, pipeline, file_pattern):
    """Generate examples as dicts."""
    beam = tfds.core.lazy_imports.apache_beam

    def _generate_examples(file_path):
      """Read the input data out of the source files."""
      example_id = 0
      meta_examples = tf.io.read_file(file_path).numpy().decode('utf-8').split(
          '\n')[:-1]
      print(meta_examples)
      for m_example in meta_examples:
        m_example = m_example.split(' ')
        image_path = os.path.join(ORIGINAL_DATA_DIR_64, file_pattern,
                                  m_example[0], m_example[1])
        example_id += 1
        yield '_'.join([m_example[0], m_example[1],
                        str(example_id)]), {
                            'image': image_path,
                            'label': int(m_example[3]),
                        }

    meta_file_pathes = tf.io.gfile.glob(
        os.path.join(ORIGINAL_DATA_DIR_64, file_pattern, 'metadata/*.npy'))
    print(len(meta_file_pathes))
    return (pipeline
            | 'Create' >> beam.Create(meta_file_pathes)
            | 'Generate' >> beam.ParDo(_generate_examples))


class Pathfinder128(tfds.core.BeamBasedBuilder):
  """Pathfinder TFDS builder (where the resolution is 128).

  The data for this dataset was generated using the script in
  https://github.com/drewlinsley/pathfinder with the default parameters, while
  followings being customized:
  ```
    args.padding = 1
    args.antialias_scale = 4
    args.paddle_margin_list = [2,3]
    args.seed_distance = 20
    args.window_size = [128,128]
    args.marker_radius = 3
    args.contour_length = 14
    args.paddle_thickness = 1.5
    args.antialias_scale = 2
    args.continuity = 1.8  # from 1.8 to 0.8, with steps of 66%
    args.distractor_length = args.contour_length / 3
    args.num_distractor_snakes = 35 / args.distractor_length
    args.snake_contrast_list = [0.9]
  ```
  """

  VERSION = tfds.core.Version('1.0.0')

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=('This is a builder for pathfinder challenge dataset'),
        features=tfds.features.FeaturesDict({
            'image': tfds.features.Image(),
            'label': tfds.features.ClassLabel(num_classes=2)
        }),
        supervised_keys=('image', 'label'),
        homepage='',
        citation="""@inproceedings{
                    Kim*2020Disentangling,
                    title={Disentangling neural mechanisms for perceptual grouping},
                    author={Junkyung Kim* and Drew Linsley* and Kalpit Thakkar and Thomas Serre},
                    booktitle={International Conference on Learning Representations},
                    year={2020},
                    url={https://openreview.net/forum?id=HJxrVA4FDS}
                    }""",
    )

  def _split_generators(self, dl_manager):
    """Downloads the data and defines the splits."""

    return [
        tfds.core.SplitGenerator(
            name='easy', gen_kwargs={'file_pattern': 'curv_baseline'}),
        tfds.core.SplitGenerator(
            name='intermediate',
            gen_kwargs={'file_pattern': 'curv_contour_length_9'}),
        tfds.core.SplitGenerator(
            name='hard', gen_kwargs={'file_pattern': 'curv_contour_length_14'})
    ]

  def _build_pcollection(self, pipeline, file_pattern):
    """Generate examples as dicts."""
    beam = tfds.core.lazy_imports.apache_beam
    def _generate_examples(file_path):
      """Read the input data out of the source files."""
      example_id = 0
      meta_examples = tf.io.read_file(
          file_path).numpy().decode('utf-8').split('\n')[:-1]
      print(meta_examples)
      for m_example in meta_examples:
        m_example = m_example.split(' ')
        image_path = os.path.join(ORIGINAL_DATA_DIR_128, file_pattern,
                                  m_example[0], m_example[1])
        example_id += 1
        yield '_'.join([m_example[0], m_example[1], str(example_id)]), {
            'image': image_path,
            'label': int(m_example[3]),
        }

    meta_file_pathes = tf.io.gfile.glob(
        os.path.join(ORIGINAL_DATA_DIR_128, file_pattern, 'metadata/*.npy'))
    print(len(meta_file_pathes))
    return (
        pipeline
        | 'Create' >> beam.Create(meta_file_pathes)
        | 'Generate' >> beam.ParDo(_generate_examples)
    )


class Pathfinder256(tfds.core.BeamBasedBuilder):
  """Pathfinder TFDS builder (where the resolution is 256).

  The data for this dataset was generated using the script in
  https://github.com/drewlinsley/pathfinder with the default parameters, while
  followings being customized:
  ```
    args.antialias_scale = 4
    args.paddle_margin_list = [3]
    args.window_size = [256,256]
    args.marker_radius = 5
    args.contour_length = 14
    args.paddle_thickness = 2
    args.antialias_scale = 2
    args.continuity = 1.8
    args.distractor_length = args.contour_length / 3
    args.num_distractor_snakes = 30 / args.distractor_length
    args.snake_contrast_list = [1.0]
  ```
  """

  VERSION = tfds.core.Version('1.0.0')

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=('This is a builder for pathfinder challenge dataset'),
        features=tfds.features.FeaturesDict({
            'image': tfds.features.Image(),
            'label': tfds.features.ClassLabel(num_classes=2)
        }),
        supervised_keys=('image', 'label'),
        homepage='',
        citation="""@inproceedings{
                    Kim*2020Disentangling,
                    title={Disentangling neural mechanisms for perceptual grouping},
                    author={Junkyung Kim* and Drew Linsley* and Kalpit Thakkar and Thomas Serre},
                    booktitle={International Conference on Learning Representations},
                    year={2020},
                    url={https://openreview.net/forum?id=HJxrVA4FDS}
                    }""",
    )

  def _split_generators(self, dl_manager):
    """Downloads the data and defines the splits."""

    return [
        tfds.core.SplitGenerator(
            name='easy', gen_kwargs={'file_pattern': 'curv_baseline'}),
        tfds.core.SplitGenerator(
            name='intermediate',
            gen_kwargs={'file_pattern': 'curv_contour_length_9'}),
        tfds.core.SplitGenerator(
            name='hard', gen_kwargs={'file_pattern': 'curv_contour_length_14'})
    ]

  def _build_pcollection(self, pipeline, file_pattern):
    """Generate examples as dicts."""
    beam = tfds.core.lazy_imports.apache_beam

    def _generate_examples(file_path):
      """Read the input data out of the source files."""
      example_id = 0
      meta_examples = tf.io.read_file(file_path).numpy().decode('utf-8').split(
          '\n')[:-1]
      print(meta_examples)
      for m_example in meta_examples:
        m_example = m_example.split(' ')
        image_path = os.path.join(ORIGINAL_DATA_DIR_256, file_pattern,
                                  m_example[0], m_example[1])
        example_id += 1
        yield '_'.join([m_example[0], m_example[1],
                        str(example_id)]), {
                            'image': image_path,
                            'label': int(m_example[3]),
                        }

    meta_file_pathes = tf.io.gfile.glob(
        os.path.join(ORIGINAL_DATA_DIR_256, file_pattern, 'metadata/*.npy'))
    print(len(meta_file_pathes))
    return (pipeline
            | 'Create' >> beam.Create(meta_file_pathes)
            | 'Generate' >> beam.ParDo(_generate_examples))



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


import logging
import math
from contextlib import nullcontext
from typing import Optional

import torch

from xformers import _is_sparse_available

if _is_sparse_available:
    from ._sputnik_sparse import SparseCS

# NOTE: Could do with a better option on when to use triton and not
_use_triton = True
if _use_triton:
    try:
        from xformers.triton.softmax import softmax as triton_softmax
    except ImportError:
        logging.warning(
            "Triton is not available, some optimizations will not be enabled."
        )
        _use_triton = False


def _create_random_sparsity(matrix, sparsity, divisible_by=4):
    assert matrix.ndim == 3
    keep = torch.rand_like(matrix[0], dtype=torch.float32) > sparsity
    nonzero = torch.nonzero(keep)
    nnz = nonzero.shape[0]
    # NOTE: need to make it a multiple of 4 for sputnik
    nonzero = nonzero[: (nnz - nnz % divisible_by)]
    i, j = nonzero.unbind(1)
    output = torch.zeros_like(matrix)
    bdim = torch.arange(matrix.shape[0], device=matrix.device)[:, None]
    output[bdim, i, j] = matrix[bdim, i, j]
    return output


def _broadcast_batch(mask, batch_size):
    if mask.ndim == 3:
        return mask
    assert mask.ndim == 2

    mask = mask.coalesce()
    values = mask.values()
    indices = mask.indices()
    nnz = len(values)
    # strategy: repeat the indices and append the extra batch dimension to the indices
    indices = indices.repeat(1, batch_size)
    # now create the batch indices
    batch_indices = torch.arange(batch_size, device=indices.device)
    batch_indices = batch_indices[:, None].expand(batch_size, nnz).flatten()

    # put them together
    indices = torch.cat([batch_indices[None, :], indices], dim=0)

    # now repeat the values
    values = values.repeat(batch_size)

    size = (batch_size,) + mask.shape

    return torch.sparse_coo_tensor(indices, values, size)


def _matmul_with_mask(
    a: torch.Tensor, b: torch.Tensor, mask: Optional[torch.Tensor]
) -> torch.Tensor:
    if mask is None:
        return a @ b

    if _is_sparse_available:
        if isinstance(mask, SparseCS):
            return mask.matmul_with_mask(a, b)
        if mask.is_sparse:
            # perform broadcasting if needed
            mask = _broadcast_batch(mask, a.shape[0])
            # coalesced is not implemented for bool tensors, so need to cast
            mask = mask.to(dtype=a.dtype)  # type: ignore  # mypy is missing the catch above
        return torch.ops.xformers.matmul_with_mask(a, b, mask)

    # Non optimized codepath
    att = a @ b
    att[~mask] = float("-inf")
    return att


def _softmax(a: torch.Tensor, causal: bool = False) -> torch.Tensor:
    if _is_sparse_available and isinstance(a, SparseCS):
        return a.softmax()

    if a.is_sparse:
        return torch.sparse.softmax(a, dim=a.ndim - 1)

    if _use_triton:
        return triton_softmax(a, mask=None, causal=causal)
    else:
        return torch.softmax(a, dim=a.ndim - 1)


if _is_sparse_available:

    class SparseBMM(torch.autograd.Function):
        @staticmethod
        def forward(ctx, a, b):
            a = a.coalesce()
            r = torch.bmm(a, b)
            ctx.save_for_backward(a, b)
            return r

        @staticmethod
        def backward(ctx, grad):
            a, b = ctx.saved_tensors

            # gradients w.r.t. a
            ga = None
            if ctx.needs_input_grad[0]:
                ga = torch.ops.xformers.matmul_with_mask(grad, b.transpose(-2, -1), a)

            # gradients w.r.t. b
            gb = None
            if ctx.needs_input_grad[1]:
                gb = a.transpose(1, 2).bmm(grad)

            return ga, gb

    def _sparse_bmm(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        """
        Batch matrix multiply between a sparse matrix and a dense matrix
        """
        assert a.ndim == b.ndim == 3
        assert a.shape[0] == b.shape[0]
        assert a.shape[2] == b.shape[1]
        # pyre-fixme[16]: The `apply` method is not visible to Pyre.
        return SparseBMM.apply(a, b)


def bmm(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
    if _is_sparse_available:
        if isinstance(a, SparseCS):
            return a.spmm(b)
        if a.is_sparse:
            return _sparse_bmm(a, b)
    return a @ b


def _apply_dropout(att, dropout):
    if dropout is None:
        return att
    # Dropout chokes on sparse tensors
    if _is_sparse_available:
        if isinstance(att, SparseCS):
            values = att.values.clone()
            values = dropout(values)
            att = SparseCS.wrap(
                att.shape,
                values,
                att.row_indices,
                att.row_offsets,
                att.column_indices,
                att._transp_info,
            )
        elif att.is_sparse:
            att = att.coalesce()
            values = att.values().clone()  # protect against in-place dropout
            values = dropout(values)
            att = torch.sparse_coo_tensor(att.indices(), values, att.shape)

        return att

    # Non optimized vanilla dropout
    att = dropout(att)
    return att


def scaled_query_key_softmax(
    q: torch.Tensor,
    k: torch.Tensor,
    att_mask: Optional[torch.Tensor],
    causal: bool = False,
) -> torch.Tensor:
    # TODO assume we have (N, S, hs) instead of (B, nh, S, hs), with N = B x nh
    # this is needed due to limitations in sparse_bmm for now

    # Self-attend: (N, S, hs) x (N, hs, S) -> (N, S, S)
    q = q / math.sqrt(k.size(-1))
    att = _matmul_with_mask(q, k.transpose(-2, -1), att_mask)

    if att_mask is not None and att_mask.dtype != torch.bool:
        att = att + att_mask

    # Softmax to get the attention probabilities
    att = _softmax(att, causal=causal)
    return att


def scaled_dot_product_attention(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    att_mask: Optional[torch.Tensor],
    dropout: Optional[torch.nn.Module] = None,
    causal: bool = False,
) -> torch.Tensor:
    autocast_disabled = (
        _is_sparse_available
        and isinstance(att_mask, SparseCS)
        or (att_mask is not None and att_mask.is_sparse)
    )

    with torch.cuda.amp.autocast(enabled=False) if autocast_disabled else nullcontext():
        if autocast_disabled:
            q, k, v = q.float(), k.float(), v.float()

        att = scaled_query_key_softmax(q, k, att_mask=att_mask, causal=causal)

        #  Optional dropout, could be part of the masking in the future
        att = _apply_dropout(att, dropout)

        # Get to the predicted values, for all heads
        # y = att @ v  # (N, S, S) x (N, S, hs) -> (N, S, hs)
        y = bmm(att, v)
    return y



# Copyright 2021 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions to get dataset pipeline for the image cls tasks."""

from lra_benchmarks.data import pathfinder
import tensorflow.compat.v1 as tf
import tensorflow_datasets as tfds

AUTOTUNE = tf.data.experimental.AUTOTUNE



def get_mnist_datasets(n_devices, batch_size=256, normalize=False):
  """Get MNIST dataset splits."""
  if batch_size % n_devices:
    raise ValueError("Batch size %d isn't divided evenly by n_devices %d" %
                     (batch_size, n_devices))

  train_dataset = tfds.load('mnist', split='train[:90%]')
  val_dataset = tfds.load('mnist', split='train[90%:]')
  test_dataset = tfds.load('mnist', split='test')

  def decode(x):
    decoded = {
        'inputs': tf.cast(x['image'], dtype=tf.int32),
        'targets': x['label']
    }
    if normalize:
      decoded['inputs'] = decoded['inputs'] / 255
    return decoded

  train_dataset = train_dataset.map(decode, num_parallel_calls=AUTOTUNE)
  val_dataset = val_dataset.map(decode, num_parallel_calls=AUTOTUNE)
  test_dataset = test_dataset.map(decode, num_parallel_calls=AUTOTUNE)

  train_dataset = train_dataset.repeat()
  train_dataset = train_dataset.batch(batch_size, drop_remainder=True)
  val_dataset = val_dataset.batch(batch_size, drop_remainder=True)
  test_dataset = test_dataset.batch(batch_size, drop_remainder=True)

  train_dataset = train_dataset.shuffle(
      buffer_size=256, reshuffle_each_iteration=True)

  return train_dataset, val_dataset, test_dataset, 10, 256, (batch_size, 28, 28,
                                                             1)


def get_cifar10_datasets(n_devices, batch_size=256, normalize=False):
  """Get CIFAR-10 dataset splits."""
  if batch_size % n_devices:
    raise ValueError("Batch size %d isn't divided evenly by n_devices %d" %
                     (batch_size, n_devices))

  train_dataset = tfds.load('cifar10', split='train[:90%]')
  val_dataset = tfds.load('cifar10', split='train[90%:]')
  test_dataset = tfds.load('cifar10', split='test')

  def decode(x):
    decoded = {
        'inputs':
            tf.cast(tf.image.rgb_to_grayscale(x['image']), dtype=tf.int32),
        'targets':
            x['label']
    }
    if normalize:
      decoded['inputs'] = decoded['inputs'] / 255
    return decoded

  train_dataset = train_dataset.map(decode, num_parallel_calls=AUTOTUNE)
  val_dataset = val_dataset.map(decode, num_parallel_calls=AUTOTUNE)
  test_dataset = test_dataset.map(decode, num_parallel_calls=AUTOTUNE)

  train_dataset = train_dataset.repeat()
  train_dataset = train_dataset.batch(batch_size, drop_remainder=True)
  val_dataset = val_dataset.batch(batch_size, drop_remainder=True)
  test_dataset = test_dataset.batch(batch_size, drop_remainder=True)

  train_dataset = train_dataset.shuffle(
      buffer_size=256, reshuffle_each_iteration=True)

  return train_dataset, val_dataset, test_dataset, 10, 256, (batch_size, 32, 32,
                                                             1)


def get_pathfinder_orig_datasets(n_devices, batch_size=256, normalize=False):
  """Get Pathfinder dataset splits."""
  if batch_size % n_devices:
    raise ValueError("Batch size %d isn't divided evenly by n_devices %d" %
                     (batch_size, n_devices))

  builder = pathfinder.Pathfinder(data_dir=_PATHFINER_TFDS_PATH)

  def get_split(split):
    ds_p = builder.as_dataset(split=f'positive{split}')
    ds_n = builder.as_dataset(split=f'negetive{split}')
    ds = tf.data.experimental.sample_from_datasets([ds_p, ds_n],
                                                   weights=None,
                                                   seed=None)
    return ds

  train_dataset = get_split('[:80%]')
  val_dataset = get_split('[80%:90%]')
  test_dataset = get_split('[90%:]')

  def decode(x):
    decoded = {
        'inputs':
            tf.cast(tf.image.rgb_to_grayscale(x['image']), dtype=tf.int32),
        'targets':
            x['label']
    }
    if normalize:
      decoded['inputs'] = decoded['inputs'] / 255
    return decoded

  train_dataset = train_dataset.map(decode, num_parallel_calls=AUTOTUNE)
  val_dataset = val_dataset.map(decode, num_parallel_calls=AUTOTUNE)
  test_dataset = test_dataset.map(decode, num_parallel_calls=AUTOTUNE)

  train_dataset = train_dataset.repeat()
  train_dataset = train_dataset.batch(batch_size, drop_remainder=True)
  val_dataset = val_dataset.batch(batch_size, drop_remainder=True)
  test_dataset = test_dataset.batch(batch_size, drop_remainder=True)

  train_dataset = train_dataset.shuffle(
      buffer_size=256, reshuffle_each_iteration=True)

  return train_dataset, val_dataset, test_dataset, 2, 256, (batch_size, 300,
                                                            300, 1)


def get_pathfinder_base_datasets(n_devices,
                                 batch_size=256,
                                 resolution=32,
                                 normalize=False,
                                 split='easy'):
  """Get Pathfinder dataset splits."""
  if batch_size % n_devices:
    raise ValueError("Batch size %d isn't divided evenly by n_devices %d" %
                     (batch_size, n_devices))

  if split not in ['easy', 'intermediate', 'hard']:
    raise ValueError("split must be in ['easy', 'intermediate', 'hard'].")

  if resolution == 32:
    builder = pathfinder.Pathfinder32(data_dir=_PATHFINER_TFDS_PATH)
    inputs_shape = (batch_size, 32, 32, 1)
  elif resolution == 64:
    builder = pathfinder.Pathfinder64(data_dir=_PATHFINER_TFDS_PATH)
    inputs_shape = (batch_size, 64, 64, 1)
  elif resolution == 128:
    builder = pathfinder.Pathfinder128(data_dir=_PATHFINER_TFDS_PATH)
    inputs_shape = (batch_size, 128, 128, 1)
  elif resolution == 256:
    builder = pathfinder.Pathfinder256(data_dir=_PATHFINER_TFDS_PATH)
    inputs_shape = (batch_size, 256, 256, 1)
  else:
    raise ValueError('Resolution must be in [32, 64, 128, 256].')

  def get_split(split):
    ds = builder.as_dataset(
        split=split, decoders={'image': tfds.decode.SkipDecoding()})

    # Filter out examples with empty images:
    ds = ds.filter(lambda x: tf.strings.length((x['image'])) > 0)

    return ds

  train_dataset = get_split(f'{split}[:80%]')
  val_dataset = get_split(f'{split}[80%:90%]')
  test_dataset = get_split(f'{split}[90%:]')

  def decode(x):
    decoded = {
        'inputs': tf.cast(tf.image.decode_png(x['image']), dtype=tf.int32),
        'targets': x['label']
    }
    if normalize:
      decoded['inputs'] = decoded['inputs'] / 255
    return decoded

  train_dataset = train_dataset.map(decode, num_parallel_calls=AUTOTUNE)
  val_dataset = val_dataset.map(decode, num_parallel_calls=AUTOTUNE)
  test_dataset = test_dataset.map(decode, num_parallel_calls=AUTOTUNE)

  train_dataset = train_dataset.repeat()
  train_dataset = train_dataset.batch(batch_size, drop_remainder=True)
  val_dataset = val_dataset.batch(batch_size, drop_remainder=True)
  test_dataset = test_dataset.batch(batch_size, drop_remainder=True)

  train_dataset = train_dataset.shuffle(
      buffer_size=256 * 8, reshuffle_each_iteration=True)

  return train_dataset, val_dataset, test_dataset, 2, 256, inputs_shape



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


import argparse
import json
from pathlib import Path
from typing import Any, Dict

reference_steps = {
    "image": 35176,
    "listops": 10000,
    "pathfinder32-curv_contour_length_14": 62400,
    "pathfinder32-curv_baseline": 62400,
    "pathfinder32-curv_contour_length_9": 62400,
    "text": 20000,
    "retrieval": 30000,
}

if __name__ == "__main__":
    # Get the user requests
    parser = argparse.ArgumentParser(
        "Collect results from a given batch of distributed results"
    )
    parser.add_argument("-ck", "--checkpoint_path", required=True)
    args = parser.parse_args()

    # Go through all the data in the given repo, try to find the end results
    root = Path(args.checkpoint_path)

    # - list all the mechanisms being benchmarked
    results: Dict[str, Any] = {}

    for attention in filter(lambda x: x.is_dir(), root.iterdir()):
        print(f"\nFound results for {attention.stem}")
        task_logs = attention.glob("*.log")
        results[attention.stem] = {}

        for task in filter(lambda x: "__0" in str(x), task_logs):
            task_name = task.stem.split("__")[0]
            print(f"Logs found for task: {task_name}")
            results[attention.stem][task_name] = -1

            # - collect the individual results
            with open(task, "r") as result_file:
                for line in reversed(result_file.readlines()):
                    if '"component": "test"' in line:
                        # Check that all the steps are done
                        res = json.loads(line)
                        if res["train_step_idx"] == reference_steps[task_name]:
                            results[attention.stem][task_name] = res["best_accu"]
                            print(
                                f"Final result found for {task_name}: {results[attention.stem][task_name]}"
                            )
                        else:
                            print(
                                "Current step: {}/{}. Not finished".format(
                                    res["train_step_idx"], reference_steps[task_name]
                                )
                            )
                        break

    print(f"\nCollected results: {json.dumps(results, indent=2)}")

    #  - reduction: compute the average
    tasks = set(t for v in results.values() for t in v.keys())
    # -- fill in the possible gaps
    for att in results.keys():
        for t in tasks:
            if t not in results[att].keys():
                results[att][t] = 0.0

    # -- add the average value
    for att in results.keys():
        results[att]["AVG"] = round(sum(results[att][t] for t in tasks) / len(tasks), 2)

    # - Format as an array, markdown style
    tasks_sort = sorted(
        set(t for v in results.values() for t in v.keys()), reverse=True
    )
    print(
        "{0:<20}".format("") + "".join("{0:<20}   ".format(t[:10]) for t in tasks_sort)
    )

    for att in results.keys():
        print(
            "{0:<20}".format(att)
            + "".join("{0:<20}   ".format(results[att][t]) for t in tasks_sort)
        )



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


from dataclasses import asdict, dataclass
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Union

import torch
import torch.nn as nn

from xformers.components import build_multi_head_attention
from xformers.components.feedforward import (
    FEEDFORWARD_REGISTRY,
    FeedforwardConfig,
    build_feedforward,
)
from xformers.components.positional_embedding import (
    POSITION_EMBEDDING_REGISTRY,
    PositionEmbeddingConfig,
    build_positional_embedding,
)
from xformers.utils import generate_matching_config

# NOTE: The Triton layernorm can be activated/deactivated from here
_is_triton_available = False

if _is_triton_available:
    try:
        from xformers.triton.layer_norm import FusedLayerNorm
    except ImportError as e:
        import logging

        logging.warning(
            f"Triton is not available, some optimizations will not be enabled.\n{e}"
        )
        _is_triton_available = False


def _to_tensor_list(
    inputs: Union[torch.Tensor, List[torch.Tensor]]
) -> List[torch.Tensor]:
    if not isinstance(inputs, list):
        inputs = [inputs]
    return inputs


class LayerPositionBitmask(int, Enum):
    First = 0b01
    Last = 0b10
    Default = 0b11


class LayerPosition:
    """ Bitmask to mark this layer as first, last, nothing or both"""

    def __init__(self):
        self.bitmask = LayerPositionBitmask.Default

    def is_first(self):
        return bool(self.bitmask & LayerPositionBitmask.First)

    def is_last(self):
        return bool(self.bitmask & LayerPositionBitmask.Last)

    def mark_not_first(self):
        self.bitmask &= ~LayerPositionBitmask.First

    def mark_not_last(self):
        self.bitmask &= ~LayerPositionBitmask.Last


class BlockType(str, Enum):
    Encoder = "encoder"
    Decoder = "decoder"


class LayerNormStyle(str, Enum):
    """Support different layer norm styles.
    See "On Layer Normalization in the Transformer Architecture",
    Xiong et al., https://arxiv.org/pdf/2002.04745v1.pdf
    """

    Pre = "pre"
    Post = "post"


# CREDITS: the following is inspired by FastAI's Transformer implementation
class Residual(nn.Module):
    """Object-oriented handling of the residual path"""

    def __init__(self, layer: nn.Module):
        super().__init__()
        self.layer = layer

    def forward(self, inputs: Union[torch.Tensor, List[torch.Tensor]], *args, **kwargs):
        inputs = _to_tensor_list(inputs)

        return inputs[0] + self.layer(*inputs, *args, **kwargs)


class PreNorm(nn.Module):
    """Adds LayerNorm before computing attention

    ..Note: If a list of inputs is passed, all of them get normalized"""

    def __init__(self, d_model: int, sublayer: nn.Module, use_triton: bool = True):
        super().__init__()
        if _is_triton_available and use_triton:
            self.norm: Union[nn.LayerNorm, FusedLayerNorm] = FusedLayerNorm(d_model)
        else:
            self.norm = nn.LayerNorm(d_model)

        self.sublayer = sublayer

    def forward(self, inputs: Union[torch.Tensor, List[torch.Tensor]], *args, **kwargs):
        inputs = _to_tensor_list(inputs)

        x_norm = [self.norm(x_) for x_ in inputs]
        return self.sublayer(*x_norm, *args, **kwargs)


class PostNorm(nn.Module):
    """Adds LayerNorm after computing attention"""

    def __init__(self, d_model: int, sublayer: nn.Module, use_triton: bool = True):
        super().__init__()
        if _is_triton_available and use_triton:
            self.norm: Union[nn.LayerNorm, FusedLayerNorm] = FusedLayerNorm(d_model)
        else:
            self.norm = nn.LayerNorm(d_model)

        self.sublayer = sublayer

    def forward(self, inputs: Union[torch.Tensor, List[torch.Tensor]], *args, **kwargs):
        inputs = _to_tensor_list(inputs)

        x = self.sublayer(*inputs, *args, **kwargs)
        return self.norm(x)


def _get_ln_factory(
    d_model: int,
    layer_norm_style: Optional[LayerNormStyle],
    residual: bool = True,
    use_triton: bool = True,
):
    def get_layer_wrapper(
        d_model: int,
        sublayer: nn.Module,
        layer_norm_style: Optional[LayerNormStyle],
        residual: bool = True,
    ):
        if residual:
            return (
                Residual(PreNorm(d_model, sublayer, use_triton))
                if layer_norm_style == LayerNormStyle.Pre
                else PostNorm(d_model, Residual(sublayer), use_triton)
            )
        return (
            PreNorm(d_model, sublayer, use_triton)
            if layer_norm_style == LayerNormStyle.Pre
            else PostNorm(d_model, sublayer, use_triton)
        )

    def ln_factory(sublayer: nn.Module):
        return get_layer_wrapper(d_model, sublayer, layer_norm_style, residual)

    return ln_factory


@dataclass(init=False)  # handle constructors explicitly to force type changes
class xFormerBlockConfig:
    dim_model: int
    feedforward_config: FeedforwardConfig
    position_encoding_config: Optional[PositionEmbeddingConfig]
    block_type: BlockType
    layer_norm_style: LayerNormStyle
    layer_position: LayerPosition
    use_triton: bool

    def __init__(
        self,
        dim_model: int,
        feedforward_config: Dict[str, Any],
        position_encoding_config: Optional[Dict[str, Any]],
        block_type: BlockType,
        layer_norm_style: LayerNormStyle = LayerNormStyle("post"),
        use_triton: bool = True,
    ):
        self.dim_model = dim_model
        self.block_type = block_type
        self.layer_norm_style = layer_norm_style

        # Fill in possible gaps in the config for subparts of the block
        self.feedforward_config = generate_matching_config(
            feedforward_config,
            FEEDFORWARD_REGISTRY[feedforward_config["name"]].config,
        )

        self.position_encoding_config = (
            generate_matching_config(
                position_encoding_config,
                POSITION_EMBEDDING_REGISTRY[position_encoding_config["name"]].config,
            )
            if position_encoding_config is not None
            else None
        )

        # Default is that this layer is the only one, so both first and last
        self.layer_position = LayerPosition()


@dataclass(init=False)
class xFormerEncoderConfig(xFormerBlockConfig):
    multi_head_config: Dict[str, Any]

    def __init__(
        self,
        dim_model: int,
        feedforward_config: Dict[str, Any],
        multi_head_config: Dict[str, Any],
        position_encoding_config: Optional[Dict[str, Any]] = None,
        layer_norm_style: str = "post",
        **kwargs,
    ):
        super().__init__(
            dim_model=dim_model,
            feedforward_config=feedforward_config,
            position_encoding_config=position_encoding_config,
            layer_norm_style=LayerNormStyle(layer_norm_style),
            block_type=BlockType("encoder"),
        )

        self.multi_head_config = multi_head_config


@dataclass(init=False)
class xFormerDecoderConfig(xFormerBlockConfig):
    multi_head_config_masked: Dict[str, Any]  # prior to encoder output
    multi_head_config_cross: Dict[str, Any]  # cross attention, takes encoder output

    def __init__(
        self,
        dim_model: int,
        feedforward_config: Dict[str, Any],
        multi_head_config_masked: Dict[str, Any],
        multi_head_config_cross: Dict[str, Any],
        position_encoding_config: Optional[Dict[str, Any]] = None,
        layer_norm_style: str = "post",
        **kwargs,
    ):
        super().__init__(
            dim_model=dim_model,
            feedforward_config=feedforward_config,
            position_encoding_config=position_encoding_config,
            layer_norm_style=LayerNormStyle(layer_norm_style),
            block_type=BlockType("encoder"),
        )

        self.multi_head_config_masked = multi_head_config_masked
        self.multi_head_config_cross = multi_head_config_cross


class xFormerEncoderBlock(torch.nn.Module):
    r""" A vanilla Transformer Encoder block """

    def __init__(self, config: xFormerEncoderConfig, **kwargs):
        super().__init__()

        self.reversible_f = None
        self.reversible_g = None
        self.layer_norm_style = config.layer_norm_style
        self.dim_model = config.dim_model

        # If this layer is the first one, and a pose encoding has been requested
        self.pose_encoding = (
            build_positional_embedding(asdict(config.position_encoding_config))
            if config.position_encoding_config and config.layer_position.is_first()
            else None
        )

        # mini helper, builds a LayerNorm with the right Pre/Post config, residuals, and the right dimensions
        ln_factory = _get_ln_factory(config.dim_model, config.layer_norm_style)

        self.mha = build_multi_head_attention(config.multi_head_config)
        self.feedforward = build_feedforward(asdict(config.feedforward_config))

        # Wrappers handle the different layer norm styles (pre- and post-) and the residual path
        self.wrap_att = ln_factory(self.mha)
        self.wrap_ff: Union[Residual, PostNorm] = ln_factory(self.feedforward)

        if (
            config.layer_norm_style == LayerNormStyle.Pre
            and config.layer_position.is_last()
        ):
            self.wrap_ff = PostNorm(config.dim_model, self.wrap_ff)

    @classmethod
    def from_config(cls, config: xFormerEncoderConfig):
        return cls(config)

    @staticmethod
    def get_reversible_layer(config) -> Tuple[nn.Module, nn.Module]:
        ln_factory = _get_ln_factory(
            config.dim_model, config.layer_norm_style, residual=False
        )

        mha = build_multi_head_attention(config.multi_head_config)
        feedforward = build_feedforward(asdict(config.feedforward_config))

        reversible_f = ln_factory(mha)
        reversible_g = ln_factory(feedforward)
        return reversible_f, reversible_g

    def forward(
        self,
        x: torch.Tensor,
        att_mask: Optional[torch.Tensor] = None,
        input_mask: Optional[torch.Tensor] = None,
    ):
        if self.pose_encoding:
            x = self.pose_encoding(x)

        # Handle the optional input masking, differs on Q, K, V
        if input_mask is not None:
            q = x
            k = x * input_mask.unsqueeze(-1)
            v = k
        else:
            q, k, v = x, x, x

        # Pre/Post norms and residual paths are already handled
        x = self.wrap_att(q, k, v, att_mask=att_mask)
        x = self.wrap_ff(x)

        return x


class xFormerDecoderBlock(torch.nn.Module):
    r"""A vanilla Transformer Decoder block

    ... note: this implementation is not (yet ?) reversible"""

    def __init__(self, config: xFormerDecoderConfig, **kwargs):
        super().__init__()

        # If this layer is the first one, and a pose encoding as been requested
        self.pose_encoding = (
            build_positional_embedding(config.position_encoding_config)
            if config.position_encoding_config and config.layer_position.is_first()
            else None
        )

        # mini helper, builds a LayerNorm with the right Pre/Post config and the right dimensions
        ln_factory = _get_ln_factory(config.dim_model, config.layer_norm_style)

        self.mha = build_multi_head_attention(config.multi_head_config_masked)
        self.cross_mha = build_multi_head_attention(config.multi_head_config_cross)
        self.feedforward = build_feedforward(config.feedforward_config)

        self.wrap_att = ln_factory(self.mha)
        self.wrap_cross = ln_factory(self.cross_mha)
        self.wrap_ff: Union[Residual, PostNorm] = ln_factory(self.feedforward)

        if (
            config.layer_norm_style == LayerNormStyle.Pre
            and config.layer_position.is_last()
        ):
            self.wrap_ff = PostNorm(config.dim_model, self.wrap_ff)

    @classmethod
    def from_config(cls, config: xFormerDecoderConfig):
        return cls(config)

    def forward(
        self,
        target: torch.Tensor,
        memory: torch.Tensor,
        encoder_att_mask: Optional[torch.Tensor] = None,
        decoder_att_mask: Optional[torch.Tensor] = None,
        input_mask: Optional[torch.Tensor] = None,
    ):
        if self.pose_encoding:
            target = self.pose_encoding(target)

        # Handle the optional input masking, differs on Q, K, V
        if input_mask is not None:
            target_q = target
            target_k = target * input_mask.unsqueeze(-1)
            target_v = target_k
        else:
            target_q, target_k, target_v = target, target, target

        x = self.wrap_att([target_q, target_k, target_v], att_mask=decoder_att_mask)
        x = self.wrap_cross([x, memory, memory], att_mask=encoder_att_mask)
        x = self.wrap_ff(x)

        return x



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


import logging
from enum import Enum
from typing import Optional

import torch
import triton
import triton.language as tl
from torch.cuda.amp import custom_bwd, custom_fwd

# CREDITS: This is adapted from the vanilla Triton example. See https://openai.com/blog/triton/
# and https://triton-lang.org/getting-started/tutorials/02-fused-softmax.html


_triton_registered_overflow = False
_triton_registered_warnings = False
_triton_softmax_fp16_enabled = False  # NOTE: PyTorch keeps softmax as fp32


class MaskType(str, Enum):
    ADD = "add"
    MUL = "mul"


def get_depth(*args, **_):
    return triton.next_power_of_2(args[-1])


# autotune: Triton will test out these configurations, and automatically pick the fastest one.
# heuristic: add arguments to the kernel call automatically given some heuristics. These arguments are passed in "meta"
# fmt: off
@triton.autotune(
    configs=[
        triton.Config({}, num_warps=1),
        triton.Config({}, num_warps=2),
        triton.Config({}, num_warps=4),
        triton.Config({}, num_warps=8),
        triton.Config({}, num_warps=16),
        triton.Config({}, num_warps=32),
    ],
    key=["K"],
)
@triton.heuristics(values={"depth": get_depth , "is_fp16": lambda *args, **_: args[0].dtype == torch.float16})
@triton.jit
def _softmax(
    Y, X, M,
    stride_ym, stride_yn,
    stride_xm, stride_xn,
    stride_mn,
    K,
    **meta,  # extra parameters which can be automatically filled in given some heuristics
):
    # fmt: om

    """
    Fused softmax kernel over a 3d tensor.
    The softmax is applied over the last dimension, meaning that this is equivalent to torch.softmax(tensor, dim=-1)

    Note, if the last dimension is large, say 128K elements, the kernel compile time can shot up to many minutes when
    the kernel is run for the first time.
    """

    m = tl.program_id(0)
    n = tl.program_id(1)

    # col indices
    k = tl.arange(0, meta["depth"])

    # the memory address of all the elements that we want to load can be computed as follows
    x_ptrs = X + m * stride_xm + n * stride_xn + k

    # load input data; pad out-of-bounds elements with 0
    io_mask = k < K

    # Causal - 1: skip on the loads directly
    if meta["causal"]:
        io_mask = io_mask & (k <= n)

    x = tl.load(x_ptrs, mask=io_mask, other=float("-inf"))

    # Causal - 2: enforce correctness over a couple of misloaded values
    if meta["causal"]:
        off = float("-inf")
        off = off.to(x.dtype)
        x = tl.where(k > n, off, x)

    if meta["use_mask"]:
        mask_ptrs = M + n * stride_mn + k
        add_mask = tl.load(mask_ptrs, io_mask, other=float("-inf"))
        x += add_mask

    # compute numerically-stable softmax
    z = x - tl.max(x, axis=0)

    if meta["is_fp16"]:
        # tl.exp() crashes on fp16 values
        # See https://github.com/openai/triton/issues/241
        z = z.to(tl.float32)

    num = tl.exp(z)
    denom = tl.sum(num, axis=0)

    if meta["log"]:
        y = z - tl.log(denom)
    else:
        y = num / denom

    # write back to Y.
    # we only write once, hence the "fused" softmax naming
    y_ptrs = Y + m * stride_ym + n * stride_yn + k

    # technically we could write only the lower triangular matrix in the causal case
    # but this is deemed to error prone
    tl.store(y_ptrs, y, mask=k < K)


# fmt: off
@triton.autotune(
    configs=[
        triton.Config({}, num_warps=1),
        triton.Config({}, num_warps=2),
        triton.Config({}, num_warps=4),
        triton.Config({}, num_warps=8),
        triton.Config({}, num_warps=16),
    ],
    key=["K"],
)
@triton.heuristics(values={"is_fp16": lambda *args, **_: args[0].dtype == torch.float16})
@triton.jit
def _softmax_backward(
    GradIn, GradOut, Out,
    stride_bm, stride_bn,
    stride_gm, stride_gn,
    stride_om, stride_on,
    K,
    **meta,
):
    # fmt: on

    """
    Compute the softmax gradients.
    ..Note: Not autotuning for now because this would lead to broken accumulated gradients
    """

    m = tl.program_id(0)
    n = tl.program_id(1)

    # col indices
    k = tl.arange(0, meta["depth"])

    # the memory address of all the elements that we want to load can be computed as follows
    grad_out_ptrs = GradOut + m * stride_gm + n * stride_gn + k
    out_ptrs = Out + m * stride_om + n * stride_on + k

    # load input data; pad out-of-bounds elements with 0
    io_mask = k < K

    # Causal - 1: skip on the loads directly
    if meta["causal"]:
        io_mask = io_mask & (k <= n)

    g = tl.load(grad_out_ptrs, mask=io_mask, other=float(0))
    o = tl.load(out_ptrs, mask=io_mask, other=float(0))

    # Causal - 2: enforce correctness over a couple of misloaded values
    if meta["causal"]:
        zero = float(0)
        zero = zero.to(g.dtype)
        g = tl.where(k > n, zero, g)
        o = tl.where(k > n, zero, o)

    if meta["log"]:
        s = tl.sum(g, 0)
        if meta["is_fp16"]:
            o = o.to(tl.float32)
        grad_in = g - tl.exp(o) * s
    else:
        # Step 1: Compute the intermediate sum used for the gradient
        s = tl.sum(g * o, 0)

        # Step 2: Compute the gradients
        grad_in = o * (g - s)

    # write back to the input gradients
    # technically we could write only the lower triangular matrix in the causal case
    # but this is deemed to error prone
    grad_in_ptrs = GradIn + m * stride_bm + n * stride_bn + k
    tl.store(grad_in_ptrs, grad_in, mask=k < K)


# Helper to handle the SPMD launch grid and error cases
class _softmax_triton(torch.autograd.Function):
    @staticmethod
    @custom_fwd(cast_inputs=torch.float16 if _triton_softmax_fp16_enabled else None)
    def forward(ctx, x, mask, log_outputs, causal):
        """
        Fused softmax implementation, using the Triton programming model.
        This only supports a reduction over the last dimension for now
        """

        y = torch.empty_like(x)

        assert x.ndim == 3, "This implementation only supports 3-dim tensors"
        assert y.stride(2) == 1 and x.stride(2) == 1

        # SPMD launch grid
        grid_2d = (
            x.shape[0],
            x.shape[1],
        )

        # enqueue GPU kernel
        use_mask = True
        if mask is None:
            #  placeholder, will not be used
            mask = x
            use_mask = False
        else:
            # Make sure that the mask is binary
            assert mask.dtype == x.dtype, "An additive mask is requested"

        _softmax[grid_2d](
            y, x, mask,
            y.stride(0), y.stride(1),
            x.stride(0), x.stride(1),
            mask.stride(0),
            x.shape[2],
            log=log_outputs,
            use_mask=use_mask,
            causal=causal
        )

        ctx.save_for_backward(y)
        ctx.log_outputs = log_outputs
        ctx.causal = causal
        return y

    @staticmethod
    @custom_bwd
    def backward(ctx, grad_out):
        (out,) = ctx.saved_tensors

        assert out.ndim == 3, "This implementation only supports 3-dim tensors"

        # SPMD launch grid
        grid_2d = (
            grad_out.shape[0],
            grad_out.shape[1],
        )

        depth = triton.next_power_of_2(out.shape[2])
        grad_in = torch.empty_like(out)  # torch.zeros is measurably slower, we'll zero out in the kernel

        assert grad_in.stride(2) == 1 and grad_out.stride(2) == 1 and out.stride(2) == 1

        # fmt: off
        _softmax_backward[grid_2d](
            grad_in, grad_out, out,
            grad_in.stride(0), grad_in.stride(1),
            grad_out.stride(0), grad_out.stride(1),
            out.stride(0), out.stride(1),
            out.shape[2],
            depth=depth,
            log=ctx.log_outputs,
            causal=ctx.causal
        )
        # fmt: on
        return grad_in, None, None, None


def softmax(x: torch.Tensor, mask: Optional[torch.Tensor] = None, causal: bool = False) -> torch.Tensor:
    r"""Applies the Softmax function to an 3-dimensional input Tensor
    rescaling them so that the elements of the n-dimensional output Tensor
    lie in the range [0,1] and sum to 1.

    Softmax is defined as:

    .. math::
        \text{Softmax}(x_{i}) = \frac{\exp(x_i)}{\sum_j \exp(x_j)}

    .. warning: softmax is computed on the last dimension of the input tensor.


    Args:
        x: input tensor.
        mask: optional mask, its application will be fused to the softmax computation if triton is used
        causal: optional performance optimization, if triton is used and the attention is causal

    Returns:
        a Tensor of the same dimension and shape as the input with
        values in the range [0, 1] and sum to 1
    """
    return _softmax_dispatch(x, log=False, mask=mask, causal=causal)


def log_softmax(x: torch.Tensor, mask: Optional[torch.Tensor] = None, causal: bool = False) -> torch.Tensor:
    r"""Applies the :math:`\log(\text{Softmax}(x))` function to an 3-dimensional
    input Tensor. The LogSoftmax formulation can be simplified as:

    .. math::
        \text{LogSoftmax}(x_{i}) = \log\left(\frac{\exp(x_i) }{ \sum_j \exp(x_j)} \right)

    Args:
        x: input tensor.

    Returns:
        a Tensor of the same dimension and shape as the input with
        values in the range [-inf, 0)
    """
    return _softmax_dispatch(x, log=True, mask=mask, causal=causal)


def _softmax_dispatch(x: torch.Tensor, log: bool, mask: Optional[torch.Tensor], causal: bool = False) -> torch.Tensor:
    # Triton is used if
    # - CUDA
    # - there's enough data to make it faster than pytorch. This could change over time, Triton is improving
    # - there was no previous failure

    global _triton_registered_overflow
    global _triton_registered_warnings

    try:
        if (
            torch.cuda.is_available()
            and x.is_cuda
            and not _triton_registered_overflow
        ):
            return _softmax_triton.apply(x, mask, log, causal)
    except triton.code_gen.OutOfResources:
        # Catch cases where the current GPU does not have enough registers to hold a full tensor line
        # fallback to PyTorch's implementation, which streams the tensor in and out
        _triton_registered_overflow = True
        logging.warning(
            "Triton softmax kernel register spillover caught."
            "Deactivating this kernel, please file an issue int the xFormers repository"
        )

    if causal and not _triton_registered_warnings:
        logging.warning(
            "Triton softmax could not be used. \
                The causal flags is being passed but it does not provide any benefit with PyTorch softmax."
        )
        _triton_registered_warnings = True

    if mask is not None:
        x += mask

    return torch.softmax(x, dim=-1)



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Optional

import torch
import triton

_gpu_is_old: Optional[bool] = None


def gpu_capabilities_older_than_70() -> bool:
    """Return True if the GPU's compute capability is older than SM70."""
    global _gpu_is_old
    if _gpu_is_old is None:
        for i in range(torch.cuda.device_count()):
            major, _ = torch.cuda.get_device_capability(f"cuda:{i}")
            if major < 7:
                _gpu_is_old = True
        if _gpu_is_old is None:
            _gpu_is_old = False
    return _gpu_is_old


SUPPORTED_CUDA_DEVICES = ["V100", "A100"]


def get_current_cuda_device():
    current_device = str(torch.cuda.get_device_properties(torch.cuda.current_device()))
    for device_str in SUPPORTED_CUDA_DEVICES:
        if current_device.find(device_str) > 0:
            return device_str

    logging.warning("Unsupported device, Triton code generation may fail")
    return "P100"  # default to an old GPU


def assert_almost_equal(x, y, decimal=2, err_msg=""):
    import numpy.testing as npt

    if isinstance(x, torch.Tensor):
        x = x.cpu().detach().numpy()
    if isinstance(y, torch.Tensor):
        y = y.cpu().detach().numpy()
    npt.assert_array_almost_equal(x, y, err_msg=err_msg, decimal=decimal)


@triton.jit
def minimum(x, y):
    """
    Computes the element-wise minimum of :code:`x` and :code:`y`.

    :param input: the first input block
    :type input: Block
    :param other: the second input block
    :type other: Block
    """
    return triton.language.where(x < y, x, y)


@triton.jit
def maximum(x, y):
    """
    Computes the element-wise maximum of :code:`x` and :code:`y`.

    :param input: the first input block
    :type input: Block
    :param other: the second input block
    :type other: Block
    """
    return triton.language.where(x > y, x, y)


@triton.jit
def swizzle2d(i, j, size_i, size_j, size_g):
    """
    transformes indices of a row-major size_i*size_j matrix into those
    of one where indices are row major for each group of size_j rows.
    For example, for size_i = size_j = 4 and size_g = 2, it will transform
    [[0 , 1 , 2 , 3 ],
     [4 , 5 , 6 , 7 ],
     [8 , 9 , 10, 11],
     [12, 13, 14, 15]]
    into
    [[0, 2,  4 , 6 ],
     [1, 3,  5 , 7 ],
     [8, 10, 12, 14],
     [9, 11, 13, 15]]
    """
    # "unrolled index in array"
    ij = i * size_j + j
    # number of elements in `size_g` groups
    # of `size_j` columns
    size_gj = size_g * size_j
    # index of the group in which (i,j) is
    group_id = ij // size_gj
    # row-index of the first element of this group
    off_i = group_id * size_g
    # last group may have fewer rows
    size_g = minimum(size_i - off_i, size_g)
    # new row and column indices
    new_i = off_i + (ij % size_g)
    new_j = (ij % size_gj) // size_g
    return new_i, new_j
